package com.vmp.demo.config;

import lombok.extern.slf4j.Slf4j;
import org.apache.http.client.HttpClient;
import org.apache.http.client.HttpRequestRetryHandler;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.impl.client.DefaultConnectionKeepAliveStrategy;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.client.LaxRedirectStrategy;
import org.apache.http.impl.conn.BasicHttpClientConnectionManager;
import org.apache.http.protocol.HttpContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.env.Environment;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.web.client.RestTemplate;

import javax.annotation.Resource;
import java.io.IOException;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.nio.charset.Charset;
import java.util.List;

@Slf4j
@Configuration
public class HttpClientConfig {

    @Resource
    private Environment env;

    @Bean
    public RestTemplate restTemplate() {

        RestTemplate restTemplate = new RestTemplate();

        HttpComponentsClientHttpRequestFactory httpRequestFactory = new HttpComponentsClientHttpRequestFactory();
        httpRequestFactory.setHttpClient(createHttpClient());
        restTemplate.setRequestFactory(httpRequestFactory);
        List<HttpMessageConverter<?>> converters = restTemplate.getMessageConverters();
        for (HttpMessageConverter<?> converter : converters) {
            if (converter instanceof StringHttpMessageConverter) {
                ((StringHttpMessageConverter) converter).setDefaultCharset(Charset.forName("UTF-8"));
            }
        }
        
        return restTemplate;
    }

    private HttpClient createHttpClient() {

        int maxRetryCount = env.getProperty(
                "httpClient.config.maxRetryCount", Integer.class, 3);
        int connectionTimeoutMs = env.getProperty(
                "httpClient.config.connectionTimeoutMs", Integer.class, 5000);
        int maxConnPerRoute = env.getProperty(
                "httpClient.config.maxConnPerRoute", Integer.class, 10);
        int maxConnTotal = env.getProperty(
                "httpClient.config.maxConnTotal", Integer.class, 100);

        RequestConfig requestConfig = RequestConfig.custom()
                .setConnectTimeout(connectionTimeoutMs)
                .setConnectionRequestTimeout(connectionTimeoutMs)
                .setSocketTimeout(connectionTimeoutMs)
                .build();

        HttpRequestRetryHandler retryHandler = new HttpRequestRetryHandler() {
            @Override
            public boolean retryRequest(IOException exception, int executionCount, HttpContext context) {
                if (executionCount >= maxRetryCount) {
                    // Do not retry if over max retry count
                    return false;
                }
                if (exception instanceof SocketTimeoutException 
                        || exception instanceof SocketException) {
                    // Retry if the connection timed out
                    exception.printStackTrace();
                    log.error(exception.getMessage());
                    return true;
                }
                return false;
            }
        };

        return HttpClients.custom()
                .setConnectionManager(new BasicHttpClientConnectionManager())
                .setMaxConnPerRoute(maxConnPerRoute)
                .setMaxConnTotal(maxConnTotal)
                .setRetryHandler(retryHandler)
                .setKeepAliveStrategy(DefaultConnectionKeepAliveStrategy.INSTANCE)
                .setRedirectStrategy(LaxRedirectStrategy.INSTANCE)
                .setDefaultRequestConfig(requestConfig)
                .setSSLSocketFactory(SSLConnectionSocketFactory.getSocketFactory())
                .setConnectionManagerShared(true)
                .build();
    }
}
